## casampi

casampi is a Python package that provides the [MPI](https://en.wikipedia.org/wiki/Message_Passing_Interface) parallelization infrastructure of [CASA](https://casa.nrao.edu/).
It is included in the monolithic tarballs of CASA 6 but it can also be used together with
casatools and casatasks when using them as separate Python modules in custom user Python
setups.

casampi can be used in two scenarios:
1. As a plugin for [casatasks](https://open-bitbucket.nrao.edu/projects/CASA/repos/casatasks/browse)
2. A module that can be imported in order to use the casampi parallelization functionality
   directly, if you are an advanced user, such as the ALMA/VLA data reduction pipeline.

In scenario 1, casampi adds parallelization at the CASA tasks level. if casampi is installed
together with casatasks, casatasks will see that casampi is available and enable the
parallel processing capabilities of CASA, see the parallel processing section of the CASA
docs: [Parallel Processing](https://casa.nrao.edu/casadocs-devel/stable/parallel-processing).
This includes:
1. Parallel imaging (following the parallelization approaches for cube and continuum imaging
   as implemented in tclean).
2. Multi-MS paralellization of multiple tasks, including flagdata, setjy, etc. as explained
   in the CASA docs.

In scenario 2, casampi is the MPI based infrastructure to build parallelization schemes,
such as the so-called "Tier0" parallelization of the ALMA/VLA pipelines whereby different
plotting and cleaning commands are distributed to different processes directly by the
pipeline parallelization code. Other advanced users can also build their own [MPI](https://en.wikipedia.org/wiki/Message_Passing_Interface)
parallelization schemes based on casampi.


### Module dependencies

Required:

1. [mpi4py](https://pypi.org/project/mpi4py/) (2.0 or later). casampi has been tested with the initial releases of CASA 6
   using mpi4py 3.0.2.

2. [casatools](https://open-bitbucket.nrao.edu/projects/CASA/repos/casatools/) (and dependencies therein). This is required for casatools.logsink.


casampi is normally used together with:
3. [casatasks](https://open-bitbucket.nrao.edu/projects/CASA/repos/casatasks/browse). If casatasks and casampi are installed together, casatasks will
   benefit from the additional parallelization functionality provided by casampi.
   casatasks is also required to run the full suite of casampi tests, as many of
   the tests exercise functionality of the CASA tasks.


### Installation

If you are using the [monolithic distribution of CASA](https://open-bitbucket.nrao.edu/projects/CASA/repos/casalith/browse), casampi is already included.
No additional steps are required to use casampi, and CASA can be run in MPI parallel mode
following the instructions given in the parallel processing section of the CASA docs:
[Parallel Processing](https://casa.nrao.edu/casadocs-devel/stable/parallel-processing).

If casampi is to be installed in a custom Python setup, then assuming that casatools
and casatasks are already installed, pip can be used to install casampi:

```
MYPP=/path/to/python3/with/casatools/and/casatasks
$ MPICC=/opt/casa/03/bin/mpicc pip install --target=$MYPP casampi
```
pip will retrieve an adequate version of casampi and its dependencies (mpi4py, etc.) if the
casa-pip repository is set up in the pip.conf file (explained elsewhere).
You might not need MYPP if working directly on your current default Python or inside a
virtual environment.
MPICC is the MPI compiler wrapper, which is required to compile mpi4py. In the example
command the path points to the usual location where the OpenMPI compilers are installed
from the CASA development packages. The CASA development packages are available from
https://casa.nrao.edu/download/repo/ where the package casa03-openmpi can be found (RHEL 6
and RHEL 7). This package is the simplest option to get working MPI libraries and compilers,
and is tested to work without issues with CASA releases 5 and 6. The CASA packages use
[OpenMPI](https://www.open-mpi.org/)[^1].

More in general, the MPI compiler can be any of the MPI variants supported by mpi4py. The
details of what packages and libraries are required depend on the MPI flavor chosen, and
specific multi-threading related flags might be needed[^1]. Note that the MPI compiler is
not needed to compile casampi but it is needed to compile mpi4py which will be installed as
a dependency of casampi.


### Importing and basic sanity check

From [casashell](https://open-bitbucket.nrao.edu/projects/CASA/repos/casashell/browse) or IPython / Python, a basic check as follows can be used to
make sure that casampi can be imported without issues and initializes correctly:

```
> from casampi.MPIEnvironment import MPIEnvironment
> MPIEnvironment.is_mpi_enabled
True
```
If the result is not `True` there might be issues with the MPI libraries or setup.


### Running tests

A suite of tests is distributed with the casampi package (under the submodule tests).
At the moment the tests are included in two files: test_casampi.py (full suite), and
test_casampi_subset.py (reduced subset that runs faster). The subset runs in about a
minute and a half whereas the full set may take over five to ten minutes.

The tests have to be run in MPI mode using mpirun (or mpiexec depending on the MPI
flavor). At the moment, the tests need data from the CASA development data repository,
although some of the test cases can be executed without data. If the test datasets are
available the subset of casmpi tests and the full set of tests can be run as follows:

```
$ MPIRUN=/opt/casa/03/bin/mpirun
$ $MPIRUN --n 4 python -m casampi.tests.test_casampi_subset
$ $MPIRUN --n 4 python -m casampi.tests.test_casampi
```
where `--n 4` to use 4 processes in parallel. Some examples are given below to run the
tests in a more verbose or user friendly way, using the unittest or pytest packages.

Note that in this case we use an mpirun included with the CASA developer packages. If
you are using the [monolithic distribution of CASA](https://open-bitbucket.nrao.edu/projects/CASA/repos/casalith/browse), the command mpicasa plays
the role of "MPIRUN" and provides the underlying MPI libraries and setup. See the
CASA docs [Parallel Processing](https://casa.nrao.edu/casadocs-devel/stable/parallel-processing)
for how to use mpicasa. Otherwise MPI wrappers and libraries with a particular setup
are required (see developer sections below).

The following test cases can be run without test data: `test_0_MPICommandClient`,
`test_MPICommandServer`, and `test_mpi_runtime_settings`. For example, a few test cases on
the "MPI Command Client" can be run if the casampi package is installed, without special
data requirements:
```
$MPIRUN --n 4 python -m unittest --verbose casampi.tests.test_casampi_subset.test_0_MPICommandClient
```
Or alternatively, if you prefer `pytest`:
```
$MPIRUN --n 4 python -m pytest --verbose casampi -k "test_casampi_subset and test_0_MPICommandClient"
```

Similary, `test_0_MPICommandClient` can be replaced in the example commands above with
`test_MPICommandServer` or `test_mpi_runtime_settings` to run the other tests that do not
require special data:

```
$MPIRUN --n 4 python -m unittest --verbose casampi.tests.test_casampi_subset.test_MPICommandServer
# or alternatively:
$MPIRUN --n 4 python -m pytest --verbose casampi -k "test_casampi_subset and test_MPICommandServer"
```

```
$MPIRUN --n 4 python -m unittest --verbose casampi.tests.test_casampi.test_mpi_runtime_settings
# or alternatively:
$MPIRUN --n 4 python -m pytest --verbose casampi -k "test_casampi and test_mpi_runtime_settings"
```

### For developers: building from sources

#### Building a casampi wheel from sources, with mpi4py already installed

If mpi4py is not installed, see section below. Once mpi4py is installed, casampi wheels
can be built from sources following the instructions given in this section.
A Python wheel can be built from the casampi package [sources](https://open-bitbucket.nrao.edu/projects/CASA/repos/casampi/browse) if casatools is installed. For
example if casatools has been installed from a wheel into `/path/to/python3-packages`, the
following command will produce a casampi wheel:
```
# Get casampi sources and go to the top level source directory:
$ PYTHONPATH=/path/to/python3-packages/ python ./setup.py bdist_wheel
```
Or, alternatively, using a local build of CASAtools:
```
$ PYTHONPATH=../casatools/build/lib.linux-x86_64-3.6/ python ./setup.py bdist_wheel
```

##### Running tests

The tests need to run in MPI mode, under mpirun (or mpiexec depending on the MPI flavor).
To run the tests we need to:
1. Point PYTHONPATH to where the dependencies (casatools, casatasks, etc.) are located
   unless already set up. This is similary as when building or testing [casatasks](https://open-bitbucket.nrao.edu/projects/CASA/repos/casatasks/browse).
2. Run the tests in MPI mode using mpirun.

In addition, casatasks is required to run the full set of tests. In the following example
we use the mpirun included with the casa-devel package:
```
$ MYPP=/path/to/python3/with/casatools/and/casatasks
$ MPIRUN=/opt/casa/03/bin/mpirun
PYTHONPATH=$MYPP $MPIRUN --n 4  python src/casampi/tests/test_casampi_subset.py
PYTHONPATH=$MYPP $MPIRUN --n 4  python src/casampi/tests/test_casampi.py
```

Once the package is installed (`pip install` using the wheel just generated), the test
files included with casampi (test_casampi and test_casampi_subset) can be run like
this:
```
$ PYTHONPATH=$MYPP $MPIRUN --n 4 python -m casampi.tests.test_casampi
```
(each test_*.py file can be run just as a python module).

Or, alternatively, if interested in using unittest options:
```
$ PYTHONPATH=$MYPP $MPIRUN --n 4 python -m unittest --verbose --buffer --catch --failfast casampi.tests.test_casampi
```
Or yet as another alternative the test files can be run directly:
```
$ PYTHONPATH=$MYPP $MPIRUN --n 4 python $MYPP/tests/test_casampi.py
```
Or they can also be run with pytest with a command like:
```
PYTHONPATH=$MYPP $MPIRUN --n 4 python -m pytest --verbose casampi
```

If you forget to run under MPI you will see errors like:
```
MPICommandClient::__init__::MPICommandClient::__init__::casa    MPI is not enabled. To use the MPI client this process needs to be run as an MPI program, under mpiexec, mpirun or similar)
```

#### Building an mpi4py wheel using the MPI compiler from casa-devel

This is only needed when mip4py is installed (bootstrapping, the first time casampi is going
to be built/installed in a development environment).
A simple way is to use `pip install` with the MPI compiler wrapper included in casa-devel.
This can be done as follows (following the convention, as above, that the variable MYPP
holds the path to the python packages installation):
```
$ MPICC=/opt/casa/03/bin/mpicc pip install --target=$MYPP mpi4py
```
pip will retrieve an adequate version of mpi4py if the casa-pip repository is set up in
the pip.conf file (explained elsewhere).
You might not need MYPP if working directly on the system default Python or inside a
virtual environment.

Then, to make sure that the installation is correct, the mpi4py tests can be run as explained
in the [mpi4py documentation](https://mpi4py.readthedocs.io/en/stable/install.html):
```
$ PYTHONPATH=$MYPP $MPIRUN -n 4 python -m mpi4py.bench helloworld
$ PYTHONPATH=$MYPP $MPIRUN -n 4 python -m mpi4py.bench ringtest
```

#### Building a wheel of mpi4py from sources using the mpi compiler from casa-devel

This is not needed if `pip install` is used (previous section), which is the simplest method.
But if needed for new development or testing of new features in the future, mpi4py can be
built following the [instructions](https://mpi4py.readthedocs.io/en/stable/install.html) given in its documentation:: 

1. Get mpi4py source distribution, for example from https://pypi.org/project/mpi4py/
2. The latest source distribution is, as of this writing, 3.0.2: https://bitbucket.org/mpi4py/mpi4py/downloads/mpi4py-3.0.2.tar.gz

The requirement, as above, is a mpi compiler wrapper. Then, in the top level of the mpi4py
sources a wheel can be built as follows:
```
$ MPICC=/opt/casa/03/bin/mpicc  python setup.py bdist_wheel
```
This will produce a wheel ready to be installed, printing messages like:

```
creating 'dist/mpi4py-3.0.2-cp36-cp36m-linux_x86_64.whl' and adding 'build/bdist.linux-x86_64/wheel' to it
```

Then we can install that wheel normally:
```
pip install --target=$MYPP dist/mpi4py-3.0.2-cp36-cp36m-linux_x86_64.whl
```
where MYPP, as above, holds the path to the Python packages installation directory.

Then the mpi4py tests can be run as in the previous section:
```
$ PYTHONPATH=$MYPP $MPIRUN -n 4 python -m mpi4py.bench helloworld
$ PYTHONPATH=$MYPP $MPIRUN -n 4 python -m mpi4py.bench ringtest
```


[^1]: CASA uses MPI with multithreading. This might require specific configuration and
compilation flags when building the MPI libraries. For example, the OpenMPI libraries used
for CASA development are compiled with the following options:
`./configure --enable-mpi-thread-multiple --enable-opal-multi-threads`.
